# This file is part of Radicale Server - Calendar Server
# Copyright © 2014 Jean-Marc Martins
# Copyright © 2012-2017 Guillaume Ayoub
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.
"""
SQLAlchemy storage backend.

"""
import radicale.syno.constant as constant
import uuid
from contextlib import contextmanager
from radicale.syno import utils
from radicale.syno.model import Model
from radicale.storage import BaseCollection, Item


class Collection(BaseCollection):
    """
    Collection stored in synology database backend.

    Supported principals and addressbooks only.
    """

    @classmethod
    def static_init(cls):
        db_url = cls.configuration.get("syno", "database_url")
        cls._model = Model(db_url, cls.logger)

    def is_legal(self):
        """Determine whether a collection is legal or not.
        """
        # constant.PATH_ROOT_COLLECTION should always exists
        if (not self.path or self.path == constant.PATH_ROOT_COLLECTION
            ) or (self._type == constant.TYPE_PRINCIPAL
                  and self._principal_id) or (
                      self._type is constant.TYPE_ADDRESSBOOK
                      and self._addressbook_id
                  ):
            return True
        else:
            return False

    @classmethod
    def discover(cls, path, depth="0"):
        """Discover a list of collections under the given ``path``.

        ``path`` is sanitized.

        If ``depth`` is "0", only the actual object under ``path`` is
        returned.

        If ``depth`` is anything but "0", it is considered as "1" and direct
        children are included in the result.

        The ``path`` is relative.
        """
        # path should already be santinized
        attrs = utils.get_attrs(path)

        # "/principal_uri/addressbook_uri/item_uri" is a request for item
        if len(attrs) > 2:
            href = attrs.pop()
        else:
            href = None

        collection_path = "/".join(attrs)

        cls.logger.debug(
            "[DISCOVER] - path: {0}, depth: {1}.".format(path,
                                                         depth)
        )

        collection = cls(collection_path)

        # requesting for vobject item
        if href:
            cls.logger.debug(
                "At depth 0, yield vobject item: {0}"
                .format(collection_path + "/" + href)
            )
            yield collection.get(href)
            return

        if not collection.is_legal():
            cls.logger.debug(
                """
                Illegal collection: path: {0}, principal_id: {1}, addressbook_id: {2}, collection_type: {3}.
                """.format(
                    collection.path,
                    collection._principal_id,
                    collection._addressbook_id,
                    collection._type
                )
            )
            return

        # requesting for collection
        cls.logger.debug(
            "At depth 0, yield collection: {0}".format(collection_path)
        )
        yield collection
        # depth 1
        if depth != "0":
            for _, item in collection.get_multi2(collection.list()):
                yield item

    def upload(self, href, vobject_item):
        """Upload a new or replace an existing item."""
        if self._type is not constant.TYPE_ADDRESSBOOK or not self._addressbook_id:
            self.logger.error("Only addressbook can upload vobject_item")
            return None

        self.logger.debug(
            "[UPLOAD] - collection_path: {0}, collection_type, :{1}, href: {2}, item: {3}"
            .format(self.path,
                    self._type,
                    href,
                    vobject_item)
        )

        self._model.upload_addressbook_object(
            self._principal_id,
            self._addressbook_id,
            href,
            vobject_item
        )

        return self.get(href)

    def _upload_all_vobject_items(self, vobject_items):
        """Replace all existing vobject items with ``vobject_items``.
        """
        if self._type is not constant.TYPE_ADDRESSBOOK or not self._addressbook_id:
            self.logger.error("Only addressbook can _upload_all_vobject_items")
            return

        self._model.delete_all_addressbook_objects(self._addressbook_id)

        for href, item in vobject_items.items():
            self._model.upload_addressbook_object(
                self._principal_id,
                self._addressbook_id,
                href,
                item
            )

    @classmethod
    def create_collection(cls, href, collection=None, props=None):
        """Create a collection.

        ``href`` is the sanitized path.

        If the collection already exists and neither ``collection`` nor
        ``props`` are set, this method shouldn't do anything. Otherwise the
        existing collection must be replaced.

        ``collection`` is a list of vobject components.

        ``props`` are metadata values for the collection.

        ``props["tag"]`` is the type of collection (VCALENDAR or VADDRESSBOOK).
        If the key ``tag`` is missing, it is guessed from the
        collection.

        """
        # Path should already be sanitized
        sane_path = utils.get_sane_path(href)

        cls.logger.debug(
            "[CREATE_COLLECTION] - path: {0}, vobject_components: {1}, props: {2}"
            .format(sane_path,
                    collection,
                    props)
        )

        if props:
            if props.get("tag") is not constant.ADDRESSBOOK_TAG:
                cls.logger.error(
                    """
                    No support for creating calendars.
                    """
                )
                return
            collection_type = constant.TYPE_ADDRESSBOOK
        else:
            collection_type = constant.TYPE_PRINCIPAL

        # create collection if needed
        cls._model.make_sure_path_exists(sane_path, collection_type)

        # path leads to principal
        if collection_type == constant.TYPE_PRINCIPAL:
            return cls(sane_path)

        # path leads to addressbook
        self = cls(sane_path)
        self.set_meta_all(props)
        if collection:
            vobject_items = {}
            for card in collection:
                id = str(uuid.uuid4())
                href = id + constant.VCF_SUFFIX
                vobject_items[href] = card
            self._upload_all_vobject_items(vobject_items)

        return self

    @classmethod
    @contextmanager
    def acquire_lock(cls, mode, user=None):
        """Set a context manager to lock the cls._model.

        ``mode`` must either be "r" for shared access or "w" for exclusive
        access.

        ``user`` is the name of the logged in user or empty.
        """
        if mode == "w":
            yield
            cls._model.commit()
        else:
            yield

    @classmethod
    def verify(cls):
        """Check the storage for errors."""
        pass

    def __init__(self, path):
        self.path = utils.get_sane_path(path)
        principal_uri, addressbook_uri = utils.get_principal_uri_and_addressbook_uri(
            path)

        self._principal_id = self._model.get_principal_id(principal_uri)
        if not addressbook_uri:
            self._addressbook_id = None
            self._type = constant.TYPE_PRINCIPAL
        else:
            self._addressbook_id = self._model.get_addressbook_id(
                addressbook_uri
            )
            self._type = constant.TYPE_ADDRESSBOOK if self._addressbook_id else None

        self._encoding = self.configuration.get("encoding", "stock")
        # DEPRECATED: Use ``self._encoding`` instead
        self.encoding = self._encoding

    @property
    def etag(self):
        """
        Encoded as quoted-string (see RFC 2616).

        Use addressbook_id + '_' + max(addressbook_object_revision) as etag
        """
        if not self._addressbook_id:
            self.logger.error(
                "Support etag for addressbooks only. path: {0}".format(
                    self.path
                )
            )
            return '""'

        max_revision_num = self._model.get_max_revision_num(
            self._addressbook_id
        )

        return '"{0}_{1}"'.format(self._addressbook_id, max_revision_num)

    @staticmethod
    def _get_addressbook_id_and_revision_id(token):
        try:
            # valid token should look like
            # "http://radicale.org/ns/sync/{addressbook_id}_{max_revision_id}"
            attrs = token.split(constant.SYNC_TOKEN_PREFIX)[1].split("_")

            addressbook_id = int(attrs[0])
            revision_id = int(attrs[1])
        except BaseException:
            return None, None

        return addressbook_id, revision_id

    def sync(self, old_token=None):
        """
        Get the current sync token and changed items for synchronization.

        ``old_token`` an old sync token which is used as the base of the
        delta update. If sync token is missing, all items are returned.

        Use 'addressbook_id + "_" + max(addressbook_object_revision)' as current sync_token
        """
        if not self._addressbook_id:
            self.logger.debug(
                "Support sync for addressbooks only. path: {0}".format(
                    self.path
                )
            )
            return None, None

        token = constant.SYNC_TOKEN_PREFIX + self.etag.strip("\"")
        addressbook_id, revision_id = self._get_addressbook_id_and_revision_id(
            old_token)

        self.logger.debug(
            "[SYNC] - path: {0}, addressbook_id: {1}, revision_id: {2}.".format(
                self.path,
                addressbook_id,
                revision_id
            )
        )

        if (addressbook_id == self._addressbook_id) and revision_id is not None:
            self.logger.debug(
                "Getting revision_id greater than {0} for addressbook_id {1}"
                .format(revision_id,
                        addressbook_id)
            )
            modified_objects = self._model.get_modified_addressbook_objects(
                self._addressbook_id,
                revision_id
            )
            return token, modified_objects
        else:
            self.logger.debug(
                "List all addressbook object for addressbook_id {0}"
                .format(addressbook_id)
            )
            return token, self.list()

    def list(self):
        """List collection items."""
        self.logger.debug(
            "[LIST] - path: {0}, type: {1}".format(self.path,
                                                   self._type)
        )
        if self._type == constant.TYPE_PRINCIPAL:
            return self._model.list_addressbooks(self._principal_id)
        elif self._type == constant.TYPE_ADDRESSBOOK:
            return self._model.list_addressbook_objects(self._addressbook_id)

    def _get_addressbooks(self, hrefs):
        if self._type != constant.TYPE_PRINCIPAL or hrefs is None:
            self.logger.error(
                """
                Only principals has addressbooks, Unable to get addressbooks for path: {0}, hrefs: {1}.
                """.format(self.path,
                           hrefs)
            )
            return None

        self.logger.debug("Getting addressbooks for {0}...".format(self.path))

        results = {}
        for href in hrefs:
            addressbook_path = self.path + "/" + str(href)
            results[href] = self.__class__(addressbook_path)

        self.logger.debug("Addressbooks: {0}".format(results))

        return results

    def _get_addressbook_objects(self, hrefs):
        """Get addressbook objects corresponding to ``hrefs``.

        Return a dict with format
        {
                "href_1": Item_1,
                "href_2": Item_2
        }
        """
        result = {}
        for href, props in self._model.get_addressbook_objects(
                self._addressbook_id, hrefs):
            props["collection"] = self
            result[href] = Item(**props)

        self.logger.debug("Addressbook objects: {0}".format(result))
        return result

    def get(self, href):
        """
        For principal collections, fetch a single addressbook.

        For addressbook collections, fetch a single addressbook object.
        """
        self.logger.debug(
            "[GET] - path: [{0}], type: {1}, href: [{2}]".format(
                self.path,
                self._type,
                href
            )
        )
        hrefs = [href]
        if self._type == constant.TYPE_PRINCIPAL:
            return self._get_addressbooks(hrefs).get(href)
        elif self._type == constant.TYPE_ADDRESSBOOK:
            return self._get_addressbook_objects(hrefs).get(href)
        else:
            return None

    def delete(self, href=None):
        """
        Delete an item.

        When ``href`` is ``None``, delete the collection.
        """
        if self._type == constant.TYPE_PRINCIPAL and href is None:
            self.logger.error(
                """
                Unable to delete principal {0}, principals cannot be deleted.
                """.format(self.path)
            )
            return

        self.logger.debug(
            "[DELETE] - path: {0}, type: {1}, href: {2}".format(
                self.path,
                self._type,
                href
            )
        )

        if href is None:
            self._model.delete_addressbook(self._addressbook_id)
        else:
            self._model.delete_addressbook_object(
                self._principal_id,
                self._addressbook_id,
                href
            )

    def get_meta(self, key=None):
        """
        Get metadata value for collection.

        Return the value of the property ``key``. If ``key`` is ``None`` return
        a dict with all properties
        """
        # {
        # "D:displayname": "Address book 2",
        # "CR:addressbook-description": "Addressbook created from web"
        # "tag": "VADDRESSBOOK",
        # "{http://inf-it.com/ns/ab/}addressbook-color": "#d1d353ff"
        # }
        if not self._addressbook_id:
            self.logger.debug("Only addressbooks can get meta")
            return {}

        self.logger.debug(
            "[GET_META] - path: {0}, key: {1}".format(self.path,
                                                      key)
        )
        props = self._model.get_addressbook_meta(self._addressbook_id)
        self.logger.debug("addressbook metas: {0}".format(props))

        return props.get(key) if key else props

    def set_meta_all(self, props):
        """
        Set metadata values for addressbook.

        ``props`` a dict with values for properties.
        """
        if not self._addressbook_id:
            self.logger.debug("Only addressbooks can set_meta_all.")
            return

        display_name = props.get(constant.PROP_ADDRESSBOOK_DISPLAY_NAME)
        color = props.get(constant.PROP_ADDRESSBOOK_COLOR, "0xFFFFFF")
        description = props.get(constant.PROP_ADDRESSBOOK_DESCRIPTION)
        if not display_name:
            self.logger.error(
                "Insufficient infomation to set addressbook meta with props: {0}"
                .format(props)
            )
            return

        self._model.set_addressbook_meta(
            self._addressbook_id,
            display_name,
            color,
            description
        )

    @property
    def last_modified(self):
        """Get the HTTP-datetime of when the collection was modified."""
        if self._type not in [
            constant.TYPE_PRINCIPAL,
            constant.TYPE_ADDRESSBOOK
        ]:
            self.logger.error(
                "Not a valid collection. path: [{0}]".format(self.path)
            )
            return None

        if self._type == constant.TYPE_PRINCIPAL:
            collection_id = self._principal_id
        else:
            collection_id = self._addressbook_id

        return self._model.get_last_modified_time(collection_id, self._type)

    def get_multi2(self, hrefs_iter):
        """
        Fetch multiple items.

        Functionally similar to ``get``, but might bring performance benefits
        on some storages when used cleverly. It's not required to return the
        requested items in the correct order. Duplicated hrefs can be ignored.

        Returns tuples with the href and the item or None if the item doesn't
        exist.
        """
        hrefs = list(hrefs_iter)

        if self._type == constant.TYPE_PRINCIPAL:
            childs = self._get_addressbooks(hrefs)
        elif self._type == constant.TYPE_ADDRESSBOOK:
            childs = self._get_addressbook_objects(hrefs)

        for href in hrefs:
            yield (href, childs.get(href))

    @classmethod
    def cleanup(cls):
        cls._model.cleanup()
